/* 
 * Copyright (c) 2000  Takuya SHIOZAKI
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL TAKUYA SHIOZAKI BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of Takuya SHIOZAKI shall
 * not be used in advertising or otherwise to promote the sale, use or
 * other dealings in this Software without prior written authorization
 * from Takuya SHIOZAKI.
 *
 * $Id: CharClock.c,v 1.9 2000/12/11 17:36:01 tshiozak Exp $
 *
 */

#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/XawInit.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Label.h>
#include <X11/Shell.h>
#include <X11/Xmu/Drawing.h>

#include <CharClock/CharClockP.h>

static void ToggleFukidashi();
static void PopupAbout();

static XtActionsRec actions[] = {
    { "toggle_fukidashi", ToggleFukidashi },
    { "popup_about", PopupAbout },
};

#define OffOf(f) XtOffsetOf(CharClockRec,f)

static XtResource resources[] = {
    { XtNclockTime1, XtCClockTime1, XtRString, sizeof(String),
      OffOf(charclock.time1), XtRString, "%I:%M:%S%p" },
    { XtNclockTime2, XtCClockTime2, XtRString, sizeof(String),
      OffOf(charclock.time2), XtRString, "" },
    { XtNclockDate1, XtCClockDate1, XtRString, sizeof(String),
      OffOf(charclock.date1), XtRString, "%Y/%m/%d" },
    { XtNclockDate2, XtCClockDate2, XtRString, sizeof(String),
      OffOf(charclock.date2), XtRString, "(%a)" },
    { XtNclockAm, XtCClockAm, XtRString, sizeof(String),
      OffOf(charclock.am), XtRString, "AM" },
    { XtNclockPm, XtCClockPm, XtRString, sizeof(String),
      OffOf(charclock.pm), XtRString, "PM" },
    { XtNclockSun, XtCClockSun, XtRString, sizeof(String),
      OffOf(charclock.week.n.sun), XtRString, "Sun" },
    { XtNclockMon, XtCClockMon, XtRString, sizeof(String),
      OffOf(charclock.week.n.mon), XtRString, "Mon" },
    { XtNclockTue, XtCClockTue, XtRString, sizeof(String),
      OffOf(charclock.week.n.tue), XtRString, "Tue" },
    { XtNclockWed, XtCClockWed, XtRString, sizeof(String),
      OffOf(charclock.week.n.wed), XtRString, "Wed" },
    { XtNclockThu, XtCClockThu, XtRString, sizeof(String),
      OffOf(charclock.week.n.thu), XtRString, "Thu" },
    { XtNclockFri, XtCClockFri, XtRString, sizeof(String),
      OffOf(charclock.week.n.fri), XtRString, "Fri" },
    { XtNclockSat, XtCClockSat, XtRString, sizeof(String),
      OffOf(charclock.week.n.sat), XtRString, "Sat" },
    { XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
      OffOf(charclock.fore), XtRString, XtDefaultForeground },
    { XtNclockBiTmap1, XtCPixmap, XtRString, sizeof(String),
      OffOf(charclock.p.bitmap1.file), XtRString, "" },
    { XtNclockEpBiTmap01, XtCPixmap, XtRString, sizeof(String),
      OffOf(charclock.p.ep_bitmap1.file), XtRString, "" },
    { XtNclockEpBiTmap02, XtCPixmap, XtRString, sizeof(String),
      OffOf(charclock.p.ep_bitmap2.file), XtRString, "" },
    { XtNtitle1, XtCTitle1, XtRString, sizeof(String),
      OffOf(charclock.title1), XtRString, "xccda" },
    { XtNtitle2, XtCTitle2, XtRString, sizeof(String),
      OffOf(charclock.title2), XtRString, "" },
    { XtNclockEpX, XtCClockEpX, XtRInt, sizeof(int),
      OffOf(charclock.ep_x), XtRImmediate, (XtPointer)0 },
    { XtNclockEpY, XtCClockEpY, XtRInt, sizeof(int),
      OffOf(charclock.ep_y), XtRImmediate, (XtPointer)0 },
    { XtNclockEpClosedTime, XtCClockEpClosedTime, XtRInt, sizeof(int),
      OffOf(charclock.ep_closed_time), XtRImmediate, (XtPointer)2 },
    { XtNclockEpIntervalMin, XtCClockEpIntervalMin, XtRInt, sizeof(int),
      OffOf(charclock.ep_interval_min), XtRImmediate, (XtPointer)40 },
    { XtNclockEpIntervalMax, XtCClockEpIntervalMax, XtRInt, sizeof(int),
      OffOf(charclock.ep_interval_max), XtRImmediate, (XtPointer)101 },
    { XtNclockTimePeriod, XtCClockTimePeriod, XtRInt, sizeof(int),
      OffOf(charclock.time_period), XtRImmediate, (XtPointer)10 },
    { XtNhTopLeftX, XtCHTopLeftX, XtRInt, sizeof(int),
      OffOf(charclock.h_top_x), XtRImmediate, (XtPointer)0 },
    { XtNhTopLeftY, XtCHTopLeftY, XtRInt, sizeof(int),
      OffOf(charclock.h_top_y), XtRImmediate, (XtPointer)0 },
    { XtNerExtentX, XtCErExtentX, XtRInt, sizeof(int),
      OffOf(charclock.er_width), XtRImmediate, (XtPointer)0 },
    { XtNerExtentY, XtCErExtentY, XtRInt, sizeof(int),
      OffOf(charclock.er_height), XtRImmediate, (XtPointer)0 },
    { XtNmsLeftXplus, XtCMsLeftXplus, XtRInt, sizeof(int),
      OffOf(charclock.ms_x_plus), XtRImmediate, (XtPointer)0 },
    { XtNmsLeftYplus, XtCMsLeftYplus, XtRInt, sizeof(int),
      OffOf(charclock.ms_y_plus), XtRImmediate, (XtPointer)0 },
    { XtNfontSet, XtCFontSet, XtRFontSet, sizeof(XFontSet),
      OffOf(charclock.fontset), XtRString, XtDefaultFontSet },
    { XtNfukidashiTapTopLeftX, XtCFukidashiTapTopLeftX, XtRInt, sizeof(int),
      OffOf(charclock.fukidashi_x), XtRImmediate, (XtPointer)0 },
    { XtNfukidashiTapTopLeftY, XtCFukidashiTapTopLeftY, XtRInt, sizeof(int),
      OffOf(charclock.fukidashi_y), XtRImmediate, (XtPointer)0 },
    { XtNfukidashiTapExtentX, XtCFukidashiTapExtentX, XtRInt, sizeof(int),
      OffOf(charclock.fukidashi_width), XtRImmediate, (XtPointer)0 },
    { XtNfukidashiTapExtentY, XtCFukidashiTapExtentY, XtRInt, sizeof(int),
      OffOf(charclock.fukidashi_height), XtRImmediate, (XtPointer)0 },
    { XtNforceFontExtentY, XtCForceFontExtentY, XtRInt, sizeof(int),
      OffOf(charclock.force_font_extent_y), XtRImmediate, (XtPointer)0 },
};

static char translations[] =
"<Btn1Down>: toggle_fukidashi()\n"
"<Key>t: toggle_fukidashi()\n"
"<Btn2Down>: popup_about()\n"
"<Key>a: popup_about()\n"
;

static void Initialize(), Realize(), Destroy(), Redisplay();
static Boolean SetValues();

CharClockClassRec charClockClassRec = {
    { /* core fields */
    /* superclass               */      (WidgetClass) &simpleClassRec,
    /* class_name               */      "CharClock",
    /* widget_size              */      sizeof(CharClockRec),
    /* class_initialize         */      XawInitializeWidgetSet,
    /* class_part_initialize    */      NULL,
    /* class_inited             */      FALSE,
    /* initialize               */      Initialize,
    /* initialize_hook          */      NULL,
    /* realize                  */      Realize,
    /* actions                  */      actions,
    /* num_actions              */      XtNumber(actions),
    /* resources                */      resources,
    /* resource_count           */      XtNumber(resources),
    /* xrm_class                */      NULLQUARK,
    /* compress_motion          */      TRUE,
    /* compress_exposure        */      TRUE,
    /* compress_enterleave      */      TRUE,
    /* visible_interest         */      FALSE,
    /* destroy                  */      Destroy,
    /* resize                   */      NULL,
    /* expose                   */      Redisplay,
    /* set_values               */      SetValues,
    /* set_values_hook          */      NULL,
    /* set_values_almost        */      XtInheritSetValuesAlmost,
    /* get_values_hook          */      NULL,
    /* accept_focus             */      NULL,
    /* version                  */      XtVersion,
    /* callback_private         */      NULL,
    /* tm_table                 */      translations,
    /* query_geometry           */      XtInheritQueryGeometry,
    /* display_accelerator      */      XtInheritDisplayAccelerator,
    /* extension                */      NULL
    },
    { /* simple fields */
    /* change_sensitive         */      XtInheritChangeSensitive
    },
    { /* charclock fields */
    /* dummy                    */      0
    }
};

WidgetClass charClockWidgetClass = (WidgetClass) &charClockClassRec;

/*
 * Windows DIB stuffs
 */

static unsigned char *reverse = NULL;  /* swap bitorder table */

typedef struct {
    unsigned char *data;
    int ofs, size;
    int width, height;
    int inverse;
} Dib;

static uint16_t
get16le(unsigned char *p)
{
    return
        (uint16_t)p[0]|
        ((uint16_t)p[1]<<8);
}

static uint32_t
get32le(unsigned char *p)
{
    return
        (uint32_t)p[0]|
        ((uint32_t)p[1]<<8)|
        ((uint32_t)p[2]<<16)|
        ((uint32_t)p[3]<<24);
}

static int
get_dib(char *file, Dib *info)
{
    int ret = -1;
    unsigned char *p = NULL;
    long int size, hsize;
    uint32_t c0, c1;
    FILE *fp;

    info->data=NULL;
    if (!(fp=fopen(file, "r")))
        goto quit;
    if (fseek(fp, 0, SEEK_END))
        goto quit;
    if ((size=ftell(fp))<0)
        goto quit;
    if (fseek(fp, 0, SEEK_SET))
        goto quit;
    if (!(p=info->data=malloc(size)))
        goto quit;
    if (size!=fread(info->data, 1, size, fp))
        goto quit;
    fclose(fp); fp=NULL;

    info->size=size;
    /* bitmapfileheader */
    /* magic +0-+1 */
    if (*p++!='B') goto quit;
    if (*p++!='M') goto quit;
    /* size +2-+5, reserved +6-+9 */
    p+=8;
    /* offset +A-+D */
    info->ofs = get32le(p); p+=4;
    if (size<info->ofs)
        goto quit;
    /* bitmapinfoheader */
    /* header size +0-+3 */
    if ((hsize=get32le(p))<40 || info->ofs<hsize)
        goto quit;
    p+=4;
    /* width/height +4-+B */
    info->width = get32le(p); p+=4;
    info->height = get32le(p); p+=4;
    /* planes +C-+D */
    p+=2;
    /* bitcount +E-+F */
    if (get16le(p)!=1)
        goto quit;
    if (size<info->ofs+((info->width+31)/32)*4*abs(info->height))
        goto quit;
    p+=2;
    /* +10 */
    p+=hsize-0x10;
    /* palette */
    c0 = get32le(p); p+=4;
    c1 = get32le(p); p+=4;
    if (c0>c1) /*XXX*/
        info->inverse=0;
    else
        info->inverse=1;
    ret = 0;
quit:
    if (fp)
        fclose(fp);
    if (ret) {
        if (info->data)
            free(info->data);
    }

    return ret;
}

static void
free_dib(Dib *dib)
{
    if (dib) {
        if (dib->data)
            free(dib->data);
        dib->data = NULL;
    }
}

static int
dib_get_size_from_bitmap(char *file, int *width, int *height)
{
    Dib dib;
    if (!get_dib(file, &dib)) {
        *width = dib.width;
        if (dib.height>0)
            *height = dib.height;
        else
            *height = -dib.height;
        free_dib(&dib);
        return 0;
    }
    return -1;
}

static int
set_reverse()
{
    int i, j;
    if (reverse)
        return 0;
    if (!(reverse=malloc(256)))
        return -1;
    for (i=0; i<256; i++) {
        reverse[i]=0;
        for (j=0; j<8; j++)
            reverse[i]|=(i&(1<<j))?(0x80>>j):0;
    }
    return 0;
}

static int
dib_read_pixmap(CharClockWidget w, struct _bitmap *bmp)
{
    Dib dib;
    int i, j, skip, width, height;
    unsigned char *data, *p, *q;

    if (set_reverse()) {
        return -1;
    }

    if (get_dib(bmp->file, &dib)) {
        return -1;
    }

    /* packing the padding 32->8 with reverse bitorder and inverse */
    width=(dib.width+7)/8;
    skip=((dib.width+31)/32)*4 - width;
    if (!(data=malloc(width*abs(dib.height)))) {
        free_dib(&dib);
        return -1;
    }
    p=dib.data + dib.ofs;
    if (dib.height>0) {
        skip-=((dib.width+31)/32)*4*2;
        p+=((dib.width+31)/32)*4*(dib.height-1);
        height=dib.height;
    } else
        height=-dib.height;
    q=data;
    for (j=0; j<height; j++) {
        for (i=0; i<width; i++) {
            *q = reverse[*p];
            p++; q++;
        }
        p+=skip;
    }
    /* */
    bmp->pixmap=
        XCreatePixmapFromBitmapData(XtDisplay(w), XtWindow(w),
                                    data,
                                    dib.width, height,
                                    dib.inverse?
                                      w->core.background_pixel:
                                      w->charclock.fore,
                                    dib.inverse?
                                      w->charclock.fore:
                                      w->core.background_pixel,
                                    w->core.depth);
    bmp->width=dib.width; bmp->height=height;
    free(data);
    free_dib(&dib);
    return 0;
}


static int
xbm_get_size_from_bitmap(char *file, int *width, int *height)
{
    unsigned char *data;

    if (XReadBitmapFileData(file, width, height, &data, NULL, NULL)) {
        *width = *height = 0;
        return -1;
    } else
        XFree(data);
    
    return 0;
}

static int
xbm_read_pixmap(CharClockWidget w, struct _bitmap *bmp)
{
    unsigned char *data;
    int dum;

    bmp->pixmap=None;
    if (XReadBitmapFileData(bmp->file, &bmp->width, &bmp->height,
                            &data, &dum, &dum)) {
        return -1;
    }
    bmp->pixmap=
        XCreatePixmapFromBitmapData(XtDisplay(w), XtWindow(w),
                                    data, bmp->width, bmp->height,
                                    w->charclock.fore,
                                    w->core.background_pixel,
                                    w->core.depth);
    XFree(data);
    return 0;
}

static void
get_size_from_bitmap(char *file, int *width, int *height)
{
    if (dib_get_size_from_bitmap(file, width, height)) {
        xbm_get_size_from_bitmap(file, width, height);
    }
}

static void
read_pixmap(CharClockWidget w, struct _bitmap *bmp)
{
    bmp->pixmap = None;
    if ((dib_read_pixmap(w, bmp) && xbm_read_pixmap(w, bmp)) ||
        !bmp->pixmap) {
        fprintf(stderr, "reading file \"%s\" is failed.\n", bmp->file);
    }
}

static void
ok_proc(Widget gw, XtPointer c)
{
    CharClockWidget w = (CharClockWidget)c;
    XtPopdown(w->charclock.p.about.popup);
}

static void
Initialize(Widget request, Widget new, ArgList args, Cardinal *nargs)
{
    CharClockWidget w = (CharClockWidget) new;

    /* set widget extents forcibly */
    if (w->core.width<=0 || w->core.height<=0) {
        int width, height;
        get_size_from_bitmap(w->charclock.p.bitmap1.file,
                             &width, &height);
        w->core.width=width;
        w->core.height=height;
    }
        
    if (w->core.width<=0) w->core.width=1;
    if (w->core.height<=0) w->core.height=1;

    w->charclock.p.interval_id = (XtIntervalId)0;
    w->charclock.p.eye_state = EyeOpen;
    w->charclock.p.eyetimer_remain = 1;
    w->charclock.p.pop_state = PopTime;
    srand((int)time(NULL));
    w->charclock.p.about.popup =
        XtVaCreatePopupShell("about_shell", transientShellWidgetClass,
                             new, XtNtitle, "about", NULL);
    w->charclock.p.about.dialog =
        XtVaCreateManagedWidget("about", dialogWidgetClass,
                                w->charclock.p.about.popup, NULL);
    XawDialogAddButton(w->charclock.p.about.dialog, "ok",
                       (XtCallbackProc)&ok_proc, w);
                       
}

static void
draw_bg(CharClockWidget w)
{
    XClearWindow(XtDisplay(w), XtWindow(w));
    XSetFunction(XtDisplay(w), w->charclock.p.gc, GXcopy);
    XCopyArea(XtDisplay(w), w->charclock.p.bitmap1.pixmap,
              XtWindow(w), w->charclock.p.gc, 0, 0,
              w->charclock.p.bitmap1.width, w->charclock.p.bitmap1.height,
              0, 0);
}

#define DS_TOP  0
#define DS_CENTER 1
#define DS_BOTTOM  2
static void
draw_string(CharClockWidget w, Drawable d,
            int x, int y, char *mb, int baseline)
{
    int len;
    XFontSetExtents *ext = XExtentsOfFontSet(w->charclock.fontset);
    
    len=strlen(mb);
    if (baseline==DS_TOP) {
        y+=w->charclock.force_font_extent_y?
            w->charclock.force_font_extent_y:
            ext->max_ink_extent.height;
    } else if (baseline==DS_CENTER) {
        y+=(w->charclock.force_font_extent_y?
            w->charclock.force_font_extent_y:
            ext->max_ink_extent.height)/2;
    }
    
    XmbDrawString(XtDisplay(w), d, w->charclock.fontset,
                  w->charclock.p.gc, x, y, mb, len);
}

static void
convert_format(CharClockWidget w, char *in, char *out, struct tm *t)
{
    int flag = 0;
    /* XXX: it has possibility of the buffer overrun */
    /* XXX: it is not sufficient for the 7bit unit coding system */
    while (*in) {
        if (flag) {
            flag=0;
            switch (*in) {
            case 'p': /* AM/PM */
                if (t->tm_hour>11) {
                    strcpy(out, w->charclock.pm);
                } else {
                    strcpy(out, w->charclock.am);
                }
                break;
            case 'I': /* hour [1-12] */
                sprintf(out, "%d", t->tm_hour%12 + (!(t->tm_hour%12)?12:0));
                break;
            case 'M': /* minutes */
                sprintf(out, "%d", t->tm_min);
                break;
            case 'S': /* minutes */
                sprintf(out, "%d", t->tm_sec);
                break;
            case 'Y': /* year */
                sprintf(out, "%d", t->tm_year+1900);
                break;
            case 'm': /* month */
                sprintf(out, "%d", t->tm_mon+1);
                break;
            case 'd': /* day of month */
                sprintf(out, "%d", t->tm_mday);
                break;
            case 'a': /* day of week (abbrev name) */
                sprintf(out, "%s", w->charclock.week.a[t->tm_wday]);
                break;
            default:
                *out='\0';
            }
            out+=strlen(out);
        } else {
            if (*in=='%')
                flag=1;
            else {
                *out++=*in;
            }
        }
        in++;
    }
    *out='\0';
}

static void
draw_fukidashi(CharClockWidget w, int force)
{
    char buf[256]; /* XXX */
    time_t cur;
    struct tm *t;
    String l1, l2;
    int d;

    time(&cur);
    if (force || memcmp(&cur, &w->charclock.p.time_prev, sizeof(cur))) {
        t=localtime(&w->charclock.p.time_prev);
        d=t->tm_mday;
        memcpy(&w->charclock.p.time_prev, &cur, sizeof(cur));
        t=localtime(&cur);
        switch (w->charclock.p.pop_state) {
        case PopTime:
            l1=w->charclock.time1;
            l2=w->charclock.time2;
            break;
        case PopDate:
            if (d==t->tm_mday && !force)
                return;
            l1=w->charclock.date1;
            l2=w->charclock.date2;
            break;
        }

        XSetForeground(XtDisplay(w), w->charclock.p.gc,
                       w->core.background_pixel);
        XFillRectangle(XtDisplay(w), XtWindow(w), w->charclock.p.gc,
                       w->charclock.h_top_x, w->charclock.h_top_y-12,
                       w->charclock.er_width, w->charclock.er_height);
        XSetForeground(XtDisplay(w), w->charclock.p.gc, w->charclock.fore);
        /* 1 */
        convert_format(w, l1, buf, t);
        draw_string(w, XtWindow(w), w->charclock.h_top_x,
                    w->charclock.h_top_y-12, buf, DS_TOP);
        /* 2 */
        convert_format(w, l2, buf, t);
        draw_string(w, XtWindow(w),
                    w->charclock.h_top_x+w->charclock.ms_x_plus,
                    w->charclock.h_top_y+w->charclock.ms_y_plus-12,
                    buf, DS_TOP);
    }
}

static void
get_gc(CharClockWidget w)
{
    /* create new gc */
    XGCValues values;

    values.function = GXcopy;
    values.foreground = w->charclock.fore;
    values.background = w->core.background_pixel;
    values.graphics_exposures = False;
    values.font = None;
    w->charclock.p.gc =
        XCreateGC(
            XtDisplay(w), XtWindow(w),
            GCBackground | GCFunction | GCGraphicsExposures,
            &values);
}

static void
tick(XtPointer client_data, XtIntervalId *id)
{
    CharClockWidget w = (CharClockWidget)client_data;

    if (--w->charclock.p.eyetimer_remain<=0) {
        switch (w->charclock.p.eye_state) {
        case EyeOpen:
            /* to close */
            XCopyArea(XtDisplay(w), w->charclock.p.ep_bitmap2.pixmap,
                      XtWindow(w), w->charclock.p.gc, 0, 0,
                      w->charclock.p.ep_bitmap2.width,
                      w->charclock.p.ep_bitmap2.height,
                      w->charclock.ep_x,
                      w->charclock.ep_y-12);
            w->charclock.p.eyetimer_remain = w->charclock.ep_closed_time;
            w->charclock.p.eye_state = EyeClose;
            break;
        case EyeClose:
            /* to open */
            XCopyArea(XtDisplay(w), w->charclock.p.ep_bitmap1.pixmap,
                      XtWindow(w), w->charclock.p.gc, 0, 0,
                      w->charclock.p.ep_bitmap1.width,
                      w->charclock.p.ep_bitmap1.height,
                      w->charclock.ep_x,
                      w->charclock.ep_y-12);
            w->charclock.p.eyetimer_remain =
                (rand()%(w->charclock.ep_interval_max-
                         w->charclock.ep_interval_min))+
                w->charclock.ep_interval_min;
            w->charclock.p.eye_state = EyeOpen;
            break;
        }
    }
    draw_fukidashi(w, FALSE);
    w->charclock.p.interval_id =
        XtAppAddTimeOut(XtWidgetToApplicationContext((Widget) w),
                        w->charclock.time_period*10, tick, client_data);
}

static void
Realize(Widget gw, XtValueMask *mask, XSetWindowAttributes *attr)
{
    CharClockWidget w = (CharClockWidget)gw;

    *mask |= CWBitGravity;
    attr->bit_gravity = ForgetGravity;
    
    (*charClockWidgetClass->core_class.superclass->core_class.realize)
        (gw, mask, attr);

    read_pixmap(w, &w->charclock.p.bitmap1);
    read_pixmap(w, &w->charclock.p.ep_bitmap1);
    read_pixmap(w, &w->charclock.p.ep_bitmap2);
    get_gc(w);

    /* create about icon */
    w->charclock.p.about_pixmap=
        XCreatePixmap(XtDisplay(w), XtWindow(w),
                      w->charclock.p.bitmap1.width,
                      w->charclock.p.bitmap1.height,
                      w->core.depth);
    XSetFunction(XtDisplay(w), w->charclock.p.gc, GXcopy);
    XCopyArea(XtDisplay(w), w->charclock.p.bitmap1.pixmap,
              w->charclock.p.about_pixmap, w->charclock.p.gc, 0, 0,
              w->charclock.p.bitmap1.width, w->charclock.p.bitmap1.height,
              0, 0);
    if (!w->charclock.title2 || *w->charclock.title2=='\0') {
        draw_string(w, w->charclock.p.about_pixmap, w->charclock.h_top_x,
                    w->charclock.h_top_y-12+w->charclock.er_height/2,
                    w->charclock.title1, DS_CENTER);
    } else {
        /* 1 */
        draw_string(w, w->charclock.p.about_pixmap, w->charclock.h_top_x,
                    w->charclock.h_top_y-12, w->charclock.title1, DS_TOP);
        /* 2 */
        draw_string(w, w->charclock.p.about_pixmap,
                    w->charclock.h_top_x+w->charclock.ms_x_plus,
                    w->charclock.h_top_y+w->charclock.ms_y_plus-12,
                    w->charclock.title2, DS_TOP);
    }

    /* set timer */
    w->charclock.p.interval_id =
        XtAppAddTimeOut(XtWidgetToApplicationContext((Widget) w),
                        w->charclock.time_period*10, tick, (XtPointer)w);
}

static void
Destroy(Widget gw)
{
    int i;
    CharClockWidget w = (CharClockWidget)w;
    Display *dpy;

    dpy=XtDisplay(w);
    if (w->charclock.p.interval_id)
        XtRemoveTimeOut(w->charclock.p.interval_id);
    XFreeGC(dpy, w->charclock.p.gc);
    XFreePixmap(dpy, w->charclock.p.bitmap1.pixmap);
    XFreePixmap(dpy, w->charclock.p.ep_bitmap1.pixmap);
    XFreePixmap(dpy, w->charclock.p.ep_bitmap2.pixmap);
    XFreePixmap(dpy, w->charclock.p.about_pixmap);
    XtFree(w->charclock.p.bitmap1.file);
    XtFree(w->charclock.p.ep_bitmap1.file);
    XtFree(w->charclock.p.ep_bitmap2.file);
    XtFree(w->charclock.about);
    XtFree(w->charclock.time1);
    XtFree(w->charclock.time2);
    XtFree(w->charclock.am);
    XtFree(w->charclock.pm);
    XtFree(w->charclock.date1);
    XtFree(w->charclock.date2);
    for (i=0; i<7; i++)
        XtFree(w->charclock.week.a[i]);
    XFreeFontSet(dpy, w->charclock.fontset);
    XtDestroyWidget(w->charclock.p.about.dialog);
    XtDestroyWidget(w->charclock.p.about.popup);
}

static void
Redisplay(Widget gw, XEvent *ev, Region reg)
{
    CharClockWidget w=(CharClockWidget)gw;
   
    draw_bg(w);
    draw_fukidashi(w, TRUE);
}

static Boolean
SetValues(Widget gcur, Widget greq, Widget gnew, ArgList args, Cardinal *nargs)
{
    CharClockWidget cur = (CharClockWidget) gcur;
    CharClockWidget new = (CharClockWidget) gnew;
    Boolean redisplay = FALSE;

    /*XXX*/
    if (cur->charclock.fore != new->charclock.fore ||
        cur->core.background_pixel != new->core.background_pixel) {
        XFreeGC(XtDisplay(gcur), cur->charclock.p.gc);
        get_gc(new);
        redisplay = TRUE;
    }

    return redisplay;
}

static void
ToggleFukidashi(Widget gw, XEvent *ev, String *param, Cardinal *nparam)
{
    CharClockWidget w = (CharClockWidget)gw;
    if ((!ev->type==ButtonPress && !ev->type==ButtonRelease) || (
        ev->xbutton.x>=w->charclock.fukidashi_x &&
        ev->xbutton.x<=w->charclock.fukidashi_x+w->charclock.fukidashi_width &&
        ev->xbutton.y>=w->charclock.fukidashi_y &&
        ev->xbutton.y<=w->charclock.fukidashi_y+w->charclock.fukidashi_height)
        ) {
        /* toggle */
        switch (w->charclock.p.pop_state) {
        case PopTime:
            /* to date */
            w->charclock.p.pop_state = PopDate;
            break;
        case PopDate:
            /* to date */
            w->charclock.p.pop_state = PopTime;
            break;
        }

        draw_bg(w);
        draw_fukidashi(w, TRUE);
    }
}

static void
PopupAbout(Widget gw, XEvent *ev, String *param, Cardinal *nparam)
{
    CharClockWidget w = (CharClockWidget)gw;
    short x, y, width, height;

    XtVaSetValues(w->charclock.p.about.dialog,
                  XtNicon, w->charclock.p.about_pixmap,
                  NULL);
    XtPopup(w->charclock.p.about.popup, XtGrabExclusive);

    /* centering the dialog to the cursor */
    XtVaGetValues(w->charclock.p.about.popup,
                  XtNwidth, &width,
                  XtNheight, &height,
                  NULL);
    x=ev->xbutton.x_root-width/2;
    y=ev->xbutton.y_root-height/2;
    if (x+width>WidthOfScreen(XtScreen(w)))
        x=WidthOfScreen(XtScreen(w))-width;
    if (y+height>HeightOfScreen(XtScreen(w)))
        y=HeightOfScreen(XtScreen(w))-height;
    if (x<0)
        x=0;
    if (y<0)
        y=0;
    XtVaSetValues(w->charclock.p.about.popup,
                  XtNx, x,
                  XtNy, y,
                  NULL);
}


/* end of file */
