#! /usr/bin/env perl

# 
# Copyright (c) 2000  Takuya SHIOZAKI
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
# 
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL TAKUYA SHIOZAKI BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
# 
# Except as contained in this notice, the name of Takuya SHIOZAKI shall
# not be used in advertising or otherwise to promote the sale, use or
# other dealings in this Software without prior written authorization
# from Takuya SHIOZAKI.
# 
# $Id: xccdaconv.perl,v 1.4 2000/12/11 17:36:01 tshiozak Exp $
# 


use strict;
use Getopt::Std;

my ($apprestmp)=":::APPRESTMP:::";
my ($appresdir)=":::APPRESDIR:::";
my ($appres)="$appresdir/...";
my ($bmpdirbase)=":::BMPDIRBASE:::";
my ($bmpdir)="$bmpdirbase/...";
my ($bindir)=":::BINDIR:::";
my ($realbinname)="xccda";
my ($binname)="$bindir/...";
my ($ccdadir)=".";

sub usage
{
    printf STDERR
        '%s: CCDA to XCCDA converter
%s -t [apptmp] -o [appres] -B [bmpdir] -i [dir] -b [binname] <name>;
  apptmp:  application template directory (input).
           default="%s"
  appres:  application default path (output).
           default="%s"
  bmpdir:  bitmap directory (output).
           default="%s"
  dir:     ccda source directory (input)
           default="%s"
  binname: symlink name (output)
           default="%s"
  name:    clock name.
', $0, $0, $apprestmp, $appres, $bmpdir, $ccdadir;
    exit 1;
}

sub err
{
    print @_;
    exit 1;
}

my(%opts);

getopt('t=o=B=b=i=', \%opts) || usage;
usage if ($#ARGV!=0);
my ($clockname) = $ARGV[0];
$appres=$appresdir."/X".$clockname;
$apprestmp=$opts{'t'} if (defined $opts{'t'});
$appres=$opts{'o'} if (defined $opts{'o'});
$bmpdirbase=$opts{'B'} if (defined $opts{'B'});
$bmpdir=$bmpdirbase."/x".$clockname;
$binname=$bindir."/x".$clockname;
$binname=$opts{'b'} if (defined $opts{'b'});
$ccdadir=$opts{'i'} if (defined $opts{'i'});

(mkdir($bmpdirbase) || err "mkdir(\"$bmpdirbase\"): $!\n" ) if (!-d $bmpdirbase);
(mkdir($bmpdir) || err "mkdir(\"$bmpdir\"): $!\n" ) if (!-d $bmpdir);

system("ln -sf $realbinname $binname");
$? && err "";

# read project file
my ($cont) = 0;
my (%val) = ();
open(IN, "nkf -S -e $ccdadir/$clockname.rcp|")
    || err "cannot open $ccdadir/$clockname.rcp: $!\n";
$? && err "";
while (<IN>) {
    my($flnextcont)=0;
    my($nextcont)=0;
    chomp;
    s/\r$//;
    if (m/\\$/) {
        $flnextcont=1;
        $_=$`;
    }
    if ($cont) {
        # confinuing
        s/\s*\"(.*)\"\s*$/\1/;
        $$cont .= $_;
        $nextcont=$cont;
    } elsif (m/^\s*title\s*\"(.*)\"\s*$/i) {
        $val{'*title'} = $1;
        $nextcont = \$val{'*title'};
    } elsif (m/^\s*STRING\s+id\s+clockAboutMessage\s*\"(.*)\"\s*$/i) {
        $val{'*about.label'} = $1;
        $nextcont = \$val{'*about.label'};
    } elsif (m/^\s*bitmap\s+id\s+clockBiTmap1\s*\"(.*)\"\s*$/i) {
        $val{'*clockBiTmap1'} = $1;
        $nextcont = \$val{'*clockBiTmap1'};
    } elsif (m/^\s*bitmap\s+id\s+clockEpBiTmap01\s*\"(.*)\"\s*$/i) {
        $val{'*clockEpBiTmap01'} = $1;
        $nextcont = \$val{'*clockEpBiTmap01'};
    } elsif (m/^\s*bitmap\s+id\s+clockEpBiTmap02\s*\"(.*)\"\s*$/i) {
        $val{'*clockEpBiTmap02'} = $1;
        $nextcont = \$val{'*clockEpBiTmap02'};
    }
    if ($flnextcont) {
        $cont = $nextcont;
    } else {
        $cont = 0;
    }
}
close(IN);

$val{'*CharClock.title1'} = $val{'*title'};

# copy bmp files
$val{'*clockBiTmap1'} || err "*clockBiTmap1 is not found.\n";
$val{'*clockEpBiTmap01'} || err "*clockEpBiTmap01 is not found.\n";
$val{'*clockEpBiTmap02'} || err "*clockEpBiTmap02 is not found.\n";
system("cp $ccdadir/$val{'*clockBiTmap1'} $bmpdir/");
$? && err "failed: cp $ccdadir/$val{'*clockBiTmap1'} $bmpdir/\n";
$val{'*clockBiTmap1'}="$bmpdir/$val{'*clockBiTmap1'}";
system("cp $ccdadir/$val{'*clockEpBiTmap01'} $bmpdir/");
$? && err "failed: cp $ccdadir/$val{'*clockEpBiTmap01'} $bmpdir/\n";
$val{'*clockEpBiTmap01'}="$bmpdir/$val{'*clockEpBiTmap01'}";
system("cp $ccdadir/$val{'*clockEpBiTmap02'} $bmpdir/");
$? && err "failed: cp $ccdadir/$val{'*clockEpBiTmap02'} $bmpdir/\n";
$val{'*clockEpBiTmap02'}="$bmpdir/$val{'*clockEpBiTmap02'}";

# read header file
open(IN, "$ccdadir/$clockname.h") ||
    err "open(\"$ccdadir/$clockname.h\"): $!\n";
while (<IN>) {
    chomp;
    s/\r$//;
    if (m/^\#define\s+(\S+)\s+(\S+)\s*$/) {
        my ($name, $val) = ($1, $2);
        if ($name eq 'clockEpX') {
            $val{'*clockEpX'} = $val;
        } elsif ($name eq 'clockEpY') {
            $val{'*clockEpY'} = $val;
        } elsif ($name eq 'clockEpClosedTime') {
            $val{'*clockEpClosedTime'} = $val;
        } elsif ($name eq 'clockEpIntervalMin') {
            $val{'*clockEpIntervalMin'} = $val;
        } elsif ($name eq 'clockEpIntervalMax') {
            $val{'*clockEpIntervalMax'} = $val;
        } elsif ($name eq 'clockTimePeriod') {
            $val{'*clockTimePeriod'} = $val;
        } elsif ($name eq 'hTopLeftX') {
            $val{'*hTopLeftX'} = $val;
        } elsif ($name eq 'hTopLeftY') {
            $val{'*hTopLeftY'} = $val;
        } elsif ($name eq 'erExtentX') {
            $val{'*erExtentX'} = $val;
        } elsif ($name eq 'erExtentY') {
            $val{'*erExtentY'} = $val;
        } elsif ($name eq 'msLeftXplus') {
            $val{'*msLeftXplus'} = $val;
        } elsif ($name eq 'msLeftYplus') {
            $val{'*msLeftYplus'} = $val;
        } elsif ($name eq 'FukidashiTapTopLeftX') {
            $val{'*fukidashiTapTopLeftX'} = $val;
        } elsif ($name eq 'FukidashiTapTopLeftY') {
            $val{'*fukidashiTapTopLeftY'} = $val;
        } elsif ($name eq 'FukidashiTapExtentX') {
            $val{'*fukidashiTapExtentX'} = $val;
        } elsif ($name eq 'FukidashiTapExtentY') {
            $val{'*fukidashiTapExtentY'} = $val;
        }
    }
}
close(IN);

# write appres file
open(OUT, ">$appres") || err "open(\"$appres\"): $!\n";
print OUT "! AUTO GENERATED.\n";
open(IN, "$apprestmp") ||
    err "open(\"$apprestmp\"): $!\n";
while (<IN>) {
    print OUT $_;
}
close(IN);
print OUT "\n";
foreach $_ (keys %val) {
    printf OUT "%s: %s\n", $_, $val{$_};
}
close(OUT);

# end of file
